/*
 * Toolkit for Counting Active S-boxes using
 * Mixed-Integer Linear Programming (MILP).
 *
 * (c) 2012 Nicky Mouha
 *
 * Counting the number of active S-boxes for AES.
 * - Differential cryptanalysis
 * - Related-key setting
 * - Any key size (default = 256)
 * - Any number of rounds (default = 14)
 *
 * The resulting MILP program is in CPLEX lp format.
 */
#include <cstdio>
#include <cstdlib>
#include <vector>
using namespace std;

      int ROUNDS     =  14; /* number of rounds */
      int keyBits    = 256; /* number of key bits */
const int blockBits  = 128; /* number of block bits */

const int MAXKC      =   8; /* maximum key size in bits, divided by 32 */
const int MAXBC      =   8; /* maximum block size in bits, divided by 32 */
const int MAXROUNDS  =  14; /* maximum number of rounds for AES */
      int next       =   0; /* next unused state byte or key byte variable index */
      int dummy      =   0; /* next unused dummy variable index */
      int KC;
      int BC;
      
FILE* tmp;
      
static const int shifts[3][4][2] = {
 { { 0, 0 },
   { 1, 3 },
   { 2, 2 },
   { 3, 1 }
 },
 { { 0, 0 },
   { 1, 5 },
   { 2, 4 },
   { 3, 3 }
 },
 { { 0, 0 },
   { 1, 7 },
   { 3, 5 },
   { 4, 4 }
 }
};
      
#define SC	((BC - 4) >> 1)
      
vector<int> sbox_input;
vector<int> atleast_one;

/* minimize active S-boxes = keep track of S-box inputs */
int S(int x) {
  sbox_input.push_back(x);
  return x;
}

/* equations for XOR operation */
int Xor(int x, int y) {
  fprintf(tmp,"x%i + x%i + x%i - 2 d%i >= 0\n", x, y, next, dummy);
  fprintf(tmp,"d%i - x%i >= 0\n", dummy, x);
  fprintf(tmp,"d%i - x%i >= 0\n", dummy, y);
  fprintf(tmp,"d%i - x%i >= 0\n", dummy, next);
  
  dummy++;
  return next++;
}

/* calculate round keys */
void KeySchedule(int k[4][MAXKC], int keyBits, int blockBits, int W[MAXROUNDS+1][4][MAXBC]) {
  int i, j, t = 0;
  int tk[4][MAXKC];
  
	for(j = 0; j < KC; j++)
		for(i = 0; i < 4; i++)
			tk[i][j] = k[i][j];
	t = 0;
	/* copy values into round key array */
	for(j = 0; (j < KC) && (t < (ROUNDS+1)*BC); j++, t++)
		for(i = 0; i < 4; i++) W[t / BC][i][t % BC] = tk[i][j];
		
	while (t < (ROUNDS+1)*BC) { /* while not enough round key material calculated */
		/* calculate new values */
		for(j = 0; (j < 1) && (t < (ROUNDS+1)*BC); j++, t++)
			for(i = 0; i < 4; i++)
				W[t / BC][i][t % BC] = tk[i][j] = Xor(tk[i][j],S(tk[(i+1)%4][KC-1]));
		if (KC != 8)
			for(j = 1; (j < KC) && (t < (ROUNDS+1)*BC); j++, t++)
				for(i = 0; i < 4; i++) W[t / BC][i][t % BC] = tk[i][j] = Xor(tk[i][j],tk[i][j-1]);
		else {
			for(j = 1; (j < KC/2) && (t < (ROUNDS+1)*BC); j++, t++)
				for(i = 0; i < 4; i++) W[t / BC][i][t % BC] = tk[i][j] = Xor(tk[i][j],tk[i][j-1]);
			for(j = KC/2; (j < KC/2 + 1) && (t < (ROUNDS+1)*BC); j++, t++)
				for(i = 0; i < 4; i++) W[t / BC][i][t % BC] = tk[i][j] = Xor(tk[i][j],S(tk[i][j-1]));
			for(j = KC/2 + 1; (j < KC) && (t < (ROUNDS+1)*BC); j++, t++)
				for(i = 0; i < 4; i++) W[t / BC][i][t % BC] = tk[i][j] = Xor(tk[i][j],tk[i][j-1]);
		}
	}
}

static void KeyAddition(int a[4][MAXBC], int rk[4][MAXBC], int BC) {
	/* Exor corresponding text input and round key input bytes
	 */
	int i, j;
	
	for(i = 0; i < 4; i++)
   		for(j = 0; j < BC; j++) a[i][j] = Xor(a[i][j],rk[i][j]);
}

static void Substitution(int a[4][MAXBC], int BC) {
	/* Replace every byte of the input by the byte at that place
	 * in the nonlinear S-box
	 */
	int i, j;
	
	for(i = 0; i < 4; i++)
		for(j = 0; j < BC; j++) a[i][j] = S(a[i][j]);
}

static void ShiftRows(int a[4][MAXBC], int d, int BC) {
	/* Row 0 remains unchanged
	 * The other three rows are shifted a variable amount
	 */
	int tmp[MAXBC];
	int i, j;
	
	for(i = 1; i < 4; i++) {
		for(j = 0; j < BC; j++) tmp[j] = a[i][(j + shifts[SC][i][d]) % BC];
		for(j = 0; j < BC; j++) a[i][j] = tmp[j];
	}
}

static void MixColumns(int a[4][MAXBC], int BC) {
  int i,j;
  
  for(j = 0; j < BC; j++) {
    for (i = 0; i < 4; i++) fprintf(tmp,"x%i + ",a[i][j]);
    for (i = 0; i < 3; i++) fprintf(tmp,"x%i + ",next+i);
    fprintf(tmp,"x%i - 5 d%i >= 0\n",next+3,dummy);

    for(i = 0; i < 4; i++)
      fprintf(tmp,"d%i - x%i >= 0\n",dummy,a[i][j]);
    for(i = 0; i < 4; i++)
      fprintf(tmp,"d%i - x%i >= 0\n",dummy,a[i][j]=next++);
    dummy++;
  }
}

int main() {
  int a[4][MAXBC];     /* the bytes of the AES state */
  int k[4][MAXKC]; /* the bytes of the AES key */
  int rk[MAXROUNDS+1][4][MAXBC]; /* AES round keys */
  
  tmp = tmpfile();

  printf("\\ AES keysize: %i, rounds: %i\n", keyBits, ROUNDS);
  
	switch (keyBits) {
	case 128: KC = 4; break;
	case 192: KC = 6; break;
	case 256: KC = 8; break;
	default : return (-1);
	}

	switch (blockBits) {
	case 128: BC = 4; break;
	case 192: BC = 6; break;
	case 256: BC = 8; break;
	default : return (-2);
	}
  
  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < KC; j++) { /* initialize variable indices */
      atleast_one.push_back(next);
      k[i][j] = next++;
    }
  }

  KeySchedule(k, keyBits, blockBits, rk);

  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < BC; j++) { /* initialize variable indices */
      atleast_one.push_back(next);
      a[i][j] = next++;
    }
  }
 
  KeyAddition(a,rk[0],BC); 

   /* ROUNDS-1 ordinary rounds */
  for(int r = 1; r < ROUNDS; r++) {
    Substitution(a,BC);
    ShiftRows(a,0,BC);
    MixColumns(a,BC);
    KeyAddition(a,rk[r],BC);
  }

  Substitution(a,BC);
  ShiftRows(a,0,BC);
  KeyAddition(a,rk[ROUNDS],BC);
  
  /* at least one S-box must be active */
  for (size_t i = 0; i != atleast_one.size()-1; i++) {
    fprintf(tmp,"x%i + ",atleast_one.at(i));
  }
  fprintf(tmp,"x%i >= 1\n\n",atleast_one.back());

  fprintf(tmp,"Binary\n"); /* binary constraints */
  for (int i = 0; i < next; i++) fprintf(tmp,"x%i\n",i);
  for (int i = 0; i < dummy; i++) fprintf(tmp,"d%i\n",i);

  fprintf(tmp,"End\n");
  
  printf("Minimize\n"); /* print objective function */
  for (size_t i = 0; i != sbox_input.size()-1; i++) {
    printf("x%i + ",sbox_input.at(i));
  }
  printf("x%i\n\n",sbox_input.back());

  printf("Subject To\n"); /* round function constraints */

  {
    char buf[4096];
    rewind(tmp);
    while (fgets(buf, sizeof (buf), tmp) != NULL) {
      printf("%s",buf);
    }
  }
  fclose(tmp); 
  
  return 0;
}
