/*
 * Toolkit for Counting Active S-boxes using
 * Mixed-Integer Linear Programming (MILP).
 *
 * (c) 2012 Nicky Mouha, Qingju Wang
 *
 * Counting the number of active S-boxes for Enocoro128v2.
 * - Linear cryptanalysis
 * - Related-key setting
 * - Any number of rounds (default = 61)
 *
 * The resulting MILP program is in CPLEX lp format.
 */
#include <stdio.h>

const int BUFFER = 32;
const int ROUNDS = 61;

const int K128_1     =  2;
const int K128_2     =  7;
const int K128_3     = 16;
const int K128_4     = 29;
const int K128_P1    =  6;
const int K128_P2    = 15;
const int K128_P3    = 28;
const int K128_SHIFT = 31;

FILE* tmp;

void threeforked_equations(int in, int out1, int out2, int dummy){
  fprintf(tmp,"x%i + x%i + x%i -2 d%i >= 0\n", in, out1, out2,dummy);
  fprintf(tmp,"d%i - x%i >= 0\n",dummy,in);
  fprintf(tmp,"d%i - x%i >= 0\n",dummy,out1);
  fprintf(tmp,"d%i - x%i >= 0\n",dummy,out2);
}

void lfunc_equations(int in1, int in2, int out1, int out2, int dummy) {
  fprintf(tmp,"x%i + x%i + x%i + x%i - 3 d%i >= 0\n",in1,in2,out1,out2,dummy);
  fprintf(tmp,"d%i - x%i >= 0\n",dummy,in1);
  fprintf(tmp,"d%i - x%i >= 0\n",dummy,in2);
  fprintf(tmp,"d%i - x%i >= 0\n",dummy,out1);
  fprintf(tmp,"d%i - x%i >= 0\n",dummy,out2);
}

int main() {
  int i, round;
  
  int buffer[BUFFER]; /* cyclic buffer */
  int top = 0; /* offset for the cyclic buffer */
  int state[2];
  int x = 0;     /* counter for the real variables */
  int dummy = 0; /* counter for the dummy variables */
  int sbox[ROUNDS*4];
  
  tmp = tmpfile();

  for (i=0; i < BUFFER; i++) buffer[i]=x++;
  state[0] = x++;
  state[1] = x++;

  for (round=0; round < ROUNDS; round++) {

    /* update top */
    
    /* update state */
    threeforked_equations(buffer[(K128_SHIFT + top) & 0x1f],state[0],x,dummy++);
    lfunc_equations(x,sbox[4*round+1]=state[1],x+1,x+2,dummy++); x++;

    state[0]= x++;
    state[1]= x++;
    
    /* update buffer */
    threeforked_equations(buffer[(K128_1 + top) & 0x1f],sbox[4*round]=x-3,x,dummy++); x++;
    threeforked_equations(buffer[(K128_P1 + top) & 0x1f],x-1,x,dummy++);
    buffer[(K128_P1 + top) & 0x1f]=x; x++;
    
    threeforked_equations(buffer[(K128_2 + top) & 0x1f],sbox[4*round+1],x,dummy++); x++;
    threeforked_equations(buffer[(K128_P2 + top) & 0x1f],x-1,x,dummy++); 
    buffer[(K128_P2 + top) & 0x1f]=x; x++;

    threeforked_equations(buffer[(K128_3 + top) & 0x1f],sbox[4*round+2]=x-6,x,dummy++); x++;
    threeforked_equations(buffer[(K128_P3 + top) & 0x1f],x-1,x,dummy++);
    buffer[(K128_P3 + top) & 0x1f]=x; x++;

    threeforked_equations(buffer[(K128_4 + top) & 0x1f],sbox[4*round+3]=x-7,x,dummy++); x++;
   
    buffer[(K128_1 + top) & 0x1f]= x-7;
    buffer[(K128_2 + top) & 0x1f]= x-5;
    buffer[(K128_3 + top) & 0x1f]= x-3;
    buffer[(K128_4 + top) & 0x1f]= x-1;
    top = (top + K128_SHIFT) & 0x1f;
}

  for (i=0; i < BUFFER; i++) {	
    if (i == BUFFER-1) {
      fprintf(tmp,"x%i >= 1\n",i);
    } else {
      fprintf(tmp,"x%i + ",i);
    }
  }
  fprintf(tmp,"\nBinary\n");
  for (i=0; i < dummy; i++) fprintf(tmp,"d%i\n",i);
  for (i=0; i < x; i++) fprintf(tmp,"x%i\n",i);
  fprintf(tmp,"End\n");

  printf("Minimize\n");
  for (i=0; i < ROUNDS*4; i++) {
    if (i == ROUNDS * 4 - 1) {
      printf("x%i\n",sbox[i]);
    } else {
      printf("x%i + ",sbox[i]);
    }
    if (i % 15 == 14) printf("\n");
  }

  printf("\nSubject To\n");
  
  {
    char buf[4096];
    rewind(tmp);
    while (fgets(buf, sizeof (buf), tmp) != NULL) {
      printf("%s",buf);
    }
  }
  fclose(tmp); 

  return 0;
}
